/* Copyright (c) All Rights Reserved, Aberdeen University, Electronics Research Group 

 *  This is a analytical File Transfer Protocol (aFTP) program integrated with a 
 *  theoretical fade error model based on uniform CDF from ONERA 

 *  Written and created by: Martin Koyabe, koyabe@erg.abdn.ac.uk
       	      Last updated: 28th January 2001
*/


#define MAX_RECEIVERS 200000
#define MAX_PASS 1000
#define NO_OF_FADE_EVENTS 11
#define RNG 1000
#define TRUE 1
#define FALSE 0

#include <time.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <limits.h>

// #define ULONG_MAX 4294967295


double	fade[NO_OF_FADE_EVENTS];
int     j, i, k, m, n;
double  file_transfer_time, end_fade_time, cumpercentPass[MAX_RECEIVERS], rateDuplicates[MAX_RECEIVERS], goodput[MAX_RECEIVERS], throughput[MAX_RECEIVERS], bandwidth[MAX_RECEIVERS];
double  start_fade_time, fade_time, end_phase_time, lossrate[MAX_RECEIVERS];
unsigned long    rate, fileSize, MSize, nPackets, packets, tot_lost_pkts, lost_pkts, lost_packets[MAX_RECEIVERS];
unsigned long    usefulPackets[MAX_PASS][MAX_RECEIVERS], nReceivers; 
unsigned long    useful_lost_pkts, lost_pkts_first, lost_pkts_last, sumPass, maxPass=1, maxnPass[MAX_PASS], nPass[MAX_RECEIVERS], duplicates[MAX_PASS][MAX_RECEIVERS];
unsigned long    totalPass, totPass[MAX_RECEIVERS], totrecvdPackets[MAX_RECEIVERS], sumrecvdPackets, recvdPackets[MAX_PASS][MAX_RECEIVERS];
unsigned long    TotalLoss, TotalDuplicates, totDuplicates[MAX_RECEIVERS], sumDuplicates, totLoss[MAX_RECEIVERS], sumLoss;
unsigned long    loss[MAX_PASS][MAX_RECEIVERS], sumusefulPackets, totusefulPackets[MAX_RECEIVERS], TotalusefulPackets[MAX_RECEIVERS];

static int first_time = TRUE;

// Open file for data output
FILE *fd_complete = fopen("aftp_complete.tr", "a+");
FILE *fd_lossrate = fopen("aftp_lossrate.tr", "a+");
FILE *fd_goodput  = fopen("aftp_goodput.tr", "a+");
FILE *fd_duplicate = fopen("aftp_duplicate.tr","a+");
FILE *fd_dups = fopen("aftp_dups.tr","a+");
FILE *fd_loss = fopen("aftp_loss.tr","a+");
FILE *fd_bandwidth = fopen("aftp_bandwidth.tr","a+");

// Procedure min()
unsigned long min(unsigned long x, unsigned long y) 
{
  if (x < y) {
    return (x);
  } else {
    return (y);
  }    
}

// Procedure uniform()
double uniform()
{
static long s1 = (unsigned int) time (NULL);
static long s2 = (unsigned int) time (NULL);
static double factor = 1.0/2147483563.0;
register long k,z;
	k= s1 /53668;
	s1 =40014*(s1%53668)-k*12211;
	if (s1 < 0) s1 += 2147483563;
	k=s2/52774;
	s2=40692*(s2%52774)-k*3791;
	if (s2 < 0) s2 += 2147483399;

	/*
	z = abs(s1 ^ s2);
	*/
	z= (s1 - 2147483563) + s2;
	if (z < 1) z += 2147483562;
	
	return(((double)(z))*factor);
}

void usage()
{
        printf("usage: mftp [options] <value>\n");
        printf("options are:\n");
        printf("        -h<help >\n");
        printf("        -f<File Size in Bytes>\n");
        printf("        -b<Bit rate in Kbps>\n");
        printf("        -p<Packet Size in Bytes>\n");
        printf("        -r<No of Receivers>\n");
        exit(0);
}


main(int argc, char *argv[])
{

    if (argc < 2) {
       usage();
       exit(1);
    }

    argc--; argv++;
    while (argc) {
	    if (argv[0][0] != '-')
		    break;
	    switch (argv[0][1]) {

	    case 'h':
	            usage();
		    break;
	    case 'f':
		    sscanf(&argv[0][2], "%d", &fileSize);
		    printf("  File size (Bytes): %d\n", fileSize);
		    break;
	    case 'p':
		    sscanf(&argv[0][2], "%d", &MSize);
		    printf("Packet size (Bytes): %d\n", MSize);
		    break;
	    case 'b':
		    sscanf(&argv[0][2], "%d", &rate);
		    printf("   Data Rate (Kbps): %d\n", rate);
		    break;
	    case 'r':
		    sscanf(&argv[0][2], "%d", &nReceivers);
		    printf("Number of Receivers: %d\n", nReceivers);
		    break;
	    default:
		    printf("Error: invalid option\n");
		    usage();
	    }
	    argc--; argv++;
    }


// Set the SEED value based on the system time
if (first_time) {
    first_time = FALSE;
    srand((unsigned int) time (NULL));
}  

// Get the fade values
fade[0]=100.0;
fade[1]=70.0;
fade[2]=20.0;
fade[3]=10.0;
fade[4]=6.0;
fade[5]=4.0;
fade[6]=2.75;
fade[7]=2.0;
fade[8]=1.75;
fade[9]=1.5;
fade[10]=1.0;


// Compute the transfer time
file_transfer_time = ((double)fileSize * 8.0) /((double)rate*1000);

// Compute Number of packets
nPackets = (unsigned long)ceil(((double)fileSize)/(double)MSize);

printf("Scanning %d Receivers.....\n", nReceivers);
for (j = 1; j <= nReceivers; j++) {

     // initialise performance parameter
     lost_packets[j] = 0;
     duplicates[1][j] = 0;
     recvdPackets[1][j] = 0;
     usefulPackets[1][j] = 0;
     loss[1][j] = 0;
     nPass[j] = 1;
     start_fade_time = 0.0;
     end_fade_time = 0.0;
     end_phase_time = 0.0;
     lost_pkts_first = 0;
     lost_pkts_last = 0;

     // Set the end of phase 
     end_phase_time = (double)file_transfer_time;

     // Choose fade duration from uniform distribution
     char * s;
     sprintf(s, "%3.1f", 10*uniform());
     int d = atoi(s);
     fade_time = fade[d];

     // printf("fade time is: %f and end phase time is: %f\n", fade_time,end_phase_time);

     // Choose a random start time for the fade & set the end of fade time
     start_fade_time = (double)(rand()%(2*(int)(ceil(file_transfer_time))));
     end_fade_time = (start_fade_time + fade_time);
     
     // printf("start fade time is: %f end fade time is: %f\n", start_fade_time,end_fade_time);

     packets = 0;
     while (packets < nPackets) {

     // check for packets in the 1st phase
     if (nPass[j] == 1) {

       // check for lost packets when there is no fade
       if (start_fade_time >= end_phase_time) {
	 lost_packets[j] = 0;
	 duplicates[nPass[j]][j] = 0;
         recvdPackets[nPass[j]][j] = nPackets;
         usefulPackets[nPass[j]][j] = nPackets;
         loss[nPass[j]][j] = lost_packets[j];
         packets = nPackets;
       }

       // check for lost packets when fade ends in the first phase
       else if (end_fade_time < end_phase_time) {
	  lost_packets[j] = (unsigned long)((rate*1000*fade_time)/(8*MSize));
          usefulPackets[nPass[j]][j] = (nPackets - lost_packets[j]);
          recvdPackets[nPass[j]][j] = usefulPackets[nPass[j]][j];
          loss[nPass[j]][j] = lost_packets[j];
         
          nPass[j]=(nPass[j]+1);
          usefulPackets[nPass[j]][j] = lost_packets[j];
          duplicates[nPass[j]][j] = (nPackets - lost_packets[j]);
          recvdPackets[nPass[j]][j] = nPackets;
          loss[nPass[j]][j] = 0;     
	  packets = nPackets;
       }

       // check for lost packets when fade continues to second phase
       else if ((start_fade_time < end_phase_time) && (end_fade_time > end_phase_time)) {
	  lost_packets[j] = (unsigned long)((rate*1000*(end_phase_time-start_fade_time))/(8*MSize));
          lost_pkts_first = lost_packets[j];
  	
          usefulPackets[nPass[j]][j] = (nPackets - lost_packets[j]);
          recvdPackets[nPass[j]][j] = usefulPackets[nPass[j]][j];
          loss[nPass[j]][j] = lost_packets[j];
          nPass[j]=(nPass[j]+1);

  	  packets = (nPackets - lost_packets[j]);
 	  start_fade_time = end_phase_time;
	  end_phase_time = (end_phase_time + file_transfer_time);
       }
     }
     else
     {
       // check for lost packets in the subsequent phases
       if (end_fade_time > end_phase_time) {
	  lost_packets[j] = (unsigned long)((rate*1000*(end_phase_time - start_fade_time))/(8*MSize))+lost_packets[j];
  	  usefulPackets[nPass[j]][j] = 0;
          duplicates[nPass[j]][j] = 0;
          recvdPackets[nPass[j]][j] = usefulPackets[nPass[j]][j];
          loss[nPass[j]][j] =  (unsigned long)((rate*1000*(end_phase_time - start_fade_time))/(8*MSize));
          nPass[j]=(nPass[j]+1);

	  start_fade_time = end_phase_time;
	  end_phase_time = (end_phase_time + file_transfer_time);	
       }

       // check for lost packets in the second phase
       else if (end_fade_time <= end_phase_time) {
	  lost_pkts_last = (unsigned long)((rate*1000*(end_fade_time - start_fade_time))/(8*MSize));
          lost_packets[j] = lost_packets[j]+lost_pkts_last;
          tot_lost_pkts = (lost_pkts_first+lost_pkts_last);

  	  usefulPackets[nPass[j]][j] = (nPackets - lost_pkts_last);
          recvdPackets[nPass[j]][j] = usefulPackets[nPass[j]][j];
          loss[nPass[j]][j] =  lost_pkts_last;

          if (tot_lost_pkts <= nPackets) {
             duplicates[nPass[j]][j] = (nPackets - tot_lost_pkts);
          }
          else {
             duplicates[nPass[j]][j] = 0;
	  }

          nPass[j]=(nPass[j]+1);
  	  usefulPackets[nPass[j]][j] = min(lost_pkts_first,lost_pkts_last);
          duplicates[nPass[j]][j] = (nPackets - usefulPackets[nPass[j]][j]);
          recvdPackets[nPass[j]][j] = nPackets;
          loss[nPass[j]][j] =  0;

          packets = nPackets;
       }
     } // end if  
   } // end while
  
   // Get the maximum number of passes
   if (nPass[j] > maxPass) {
     maxPass = nPass[j];
   }
 } // end for
 
 // output the values of Packet Loss 

  fprintf(fd_complete, "\nMFTP Simulation data\n");
  fprintf(fd_complete, "Receivers: %d, Pkt Size: %d Bytes, File Size: %d Bytes, Rate: %d Kbps\n", nReceivers,MSize,fileSize,rate);
  fprintf(fd_complete, "Time(s) Cumulative Percent Complete\n\n");
  fprintf(fd_lossrate, "\nMFTP Simulation data\n");
  fprintf(fd_lossrate, "Receivers: %d, Pkt Size: %d Bytes, File Size: %d Bytes, Rate: %d Kbps\n", nReceivers,MSize,fileSize,rate);
  fprintf(fd_lossrate, "Time(s) Loss Rate\n\n");
  fprintf(fd_goodput, "\nMFTP Simulation data\n");
  fprintf(fd_goodput, "Receivers: %d, Pkt Size: %d Bytes, File Size: %d Bytes, Rate: %d Kbps\n", nReceivers,MSize,fileSize,rate);
  fprintf(fd_goodput, "Time(s) Goodput\n\n");
  fprintf(fd_bandwidth, "\nMFTP Simulation data\n");
  fprintf(fd_bandwidth, "Receivers: %d, Pkt Size: %d Bytes, File Size: %d Bytes, Rate: %d Kbps\n", nReceivers,MSize,fileSize,rate);
  fprintf(fd_bandwidth, "Time(s) Bandwidth\n\n");
  fprintf(fd_loss,"\nTime (s)  Loss Rate - Receivers: %d File Size: %d\n",nReceivers,fileSize);
  fprintf(fd_dups,"\nTime (s)  Total Duplicates - Receivers: %d File Size: %d\n",nReceivers,fileSize);

  totalPass = 0;
  TotalDuplicates = 0;
  TotalLoss = 0;
  for (i = 1; i <= maxPass; i++) {
      sumPass = 0;
      sumDuplicates = 0;
      sumrecvdPackets = 0;
      sumusefulPackets = 0;
      sumLoss = 0;

      for (m = 1; m <= nReceivers; m++) {
	  if (nPass[m] = i) {
            TotalusefulPackets[m] = (TotalusefulPackets[m] + usefulPackets[nPass[m]][m]);
            if (TotalusefulPackets[m] >= nPackets) {

	       sumPass = (sumPass+1);
            }
	    sumDuplicates = (sumDuplicates+duplicates[nPass[m]][m]);
            sumrecvdPackets = (sumrecvdPackets+recvdPackets[nPass[m]][m]);
	    sumusefulPackets = (sumusefulPackets+usefulPackets[nPass[m]][m]);
            sumLoss = (sumLoss+loss[nPass[m]][m]);
          }
      }

      totPass[i] = sumPass;
      totDuplicates[i] = sumDuplicates;
      totrecvdPackets[i] = sumrecvdPackets;
      totusefulPackets[i] = sumusefulPackets;
      totLoss[i] = sumLoss;

      TotalDuplicates = (TotalDuplicates + totDuplicates[i]);
      TotalLoss = (TotalLoss + totLoss[i]);

      if (totPass[i] > 0) {
         cumpercentPass[i] = ((double)(totPass[i]*100.0)/(double)(nReceivers));
	 //        printf("at time %f Cumulative Percent Pass is: %f\n",(i*file_transfer_time),(double)cumpercentPass[i]);
	 //	 printf("Percentage Complete recievers: %f\n",cumpercentPass[i]);
         fprintf(fd_complete,"%f  %f\n",(i*file_transfer_time),cumpercentPass[i]);
      }

      if (totDuplicates[i] > 0) {
         rateDuplicates[i] = ((double)(totDuplicates[i])/(double)(nReceivers));
         // printf("at time %f Duplicates Per Receivers is: %f\n",(i*file_transfer_time),(double)rateDuplicates[i]);
         fprintf(fd_duplicate,"%f  %f\n",(i*file_transfer_time),rateDuplicates[i]);
      }

      if (totusefulPackets[i] > 0) {
         goodput[i] = ((double)(totusefulPackets[i])/(double)(nReceivers*nPackets));
         // printf("at time %f Gooput is: %f\n",(i*file_transfer_time),(double)goodput[i]);
         fprintf(fd_goodput,"%f  %f\n",(i*file_transfer_time),goodput[i]);
      }

      if (totusefulPackets[i] > 0) {
         bandwidth[i] = ((double)(nReceivers*nPackets)/(double)(totusefulPackets[i]));
         // printf("at time %f Duplicates Per Receivers is: %f\n",(i*file_transfer_time),(double)rateDuplicates[i]);
         fprintf(fd_bandwidth,"%f  %f\n",(i*file_transfer_time),bandwidth[i]);
      }

      if (totLoss[i] > 0) {
         lossrate[i] = ((double)(totLoss[i])/(double)(nReceivers*nPackets));
	 //         printf("at time %f Total Loss is: %f product is: %d\n",(i*file_transfer_time),(unsigned long)TotalLoss*100.0/(unsigned long)(i*nReceivers*nPackets),(unsigned long)(i*nReceivers*nPackets));
         fprintf(fd_lossrate,"%f  %f\n",(i*file_transfer_time),lossrate[i]);
         fprintf(fd_loss,"%f  %f\n",(i*file_transfer_time),((unsigned long)TotalLoss*100.0)/(unsigned long)(i*nReceivers*nPackets));
      }
  }

  fprintf(fd_dups,"%d  %d\n",nReceivers,TotalDuplicates);
  printf("Successfully Completed...\n\n");

  fclose(fd_complete);
  fclose(fd_duplicate);
  fclose(fd_lossrate);
  fclose(fd_goodput);
  fclose(fd_dups);
  fclose(fd_loss);
  fclose(fd_bandwidth);
}





