/*
 *  Copyright (c) All Rights Reserved, Aberdeen University, Electronics Research Group 

 *  This is a analytical File Transfer Protocol (aFTP) program integrated with a 
 *  theoretical fade error model based on uniform CDF from ONERA 

 *  Written and created by: Martin Koyabe, koyabe@erg.abdn.ac.uk
       	      Last updated: 28th January 2001
*/


#define MAX_RECEIVERS 200000
#define MAX_PASS 50
#define NO_OF_FADE_EVENTS 11
#define RNG 1000
#define TRUE 1
#define FALSE 0

#include <time.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <limits.h>


double	fade[NO_OF_FADE_EVENTS];
int     j, i, k, m;
double  file_transfer_time, end_fade_time, cumpercentPass[MAX_RECEIVERS], rateDuplicates[MAX_PASS][MAX_RECEIVERS], goodput[MAX_PASS][MAX_RECEIVERS], throughput[MAX_PASS][MAX_RECEIVERS], bandwidth[MAX_PASS][MAX_RECEIVERS];
double  start_fade_time, fade_time, end_phase_time, lossrate[MAX_PASS][MAX_RECEIVERS], cumlossrate[MAX_PASS][MAX_RECEIVERS];
unsigned long    rate, fileSize, MSize, nPackets, packets, tot_lost_pkts, lost_pkts, lost_packets[MAX_RECEIVERS];
unsigned long    usefulPackets[MAX_PASS][MAX_RECEIVERS], nReceivers; 
unsigned long    lost_pkts_first, lost_pkts_last, sumPass, maxPass=1, maxnPass[MAX_PASS], nPass[MAX_RECEIVERS], duplicates[MAX_PASS][MAX_RECEIVERS];
unsigned long    totalPass, totPass[MAX_RECEIVERS], totrecvdPackets[MAX_RECEIVERS], sumrecvdPackets, recvdPackets[MAX_PASS][MAX_RECEIVERS];
unsigned long    TotalLoss, TotalDuplicates, totDuplicates[MAX_RECEIVERS], sumDuplicates, totLoss[MAX_RECEIVERS], sumLoss, loss[MAX_PASS][MAX_RECEIVERS];

static int first_time = TRUE;

// Open file for data output
FILE *fd_cumlossrate = fopen("newaftp_cumlossrate.tr", "a+");
FILE *fd_loss = fopen("newaftp_loss.tr", "a+");
FILE *fd_goodput  = fopen("newaftp_goodput.tr", "a+");
FILE *fd_throughput  = fopen("newaftp_throughput.tr", "a+");
FILE *fd_duplicate = fopen("newaftp_duplicate.tr","a+");
FILE *fd_dups = fopen("newaftp_dups.tr","a+");
FILE *fd_bandwidth = fopen("newaftp_bandwidth.tr","a+");

// Procedure min()
unsigned long min(unsigned long x, unsigned long y) 
{
  if (x < y) {
    return (x);
  } else {
    return (y);
  }    
}

// Procedure uniform()
double uniform()
{
static long s1 = (unsigned int) time (NULL);
static long s2 = (unsigned int) time (NULL);
static double factor = 1.0/2147483563.0;
register long k,z;
	k= s1 /53668;
	s1 =40014*(s1%53668)-k*12211;
	if (s1 < 0) s1 += 2147483563;
	k=s2/52774;
	s2=40692*(s2%52774)-k*3791;
	if (s2 < 0) s2 += 2147483399;

	/*
	z = abs(s1 ^ s2);
	*/
	z= (s1 - 2147483563) + s2;
	if (z < 1) z += 2147483562;
	
	return(((double)(z))*factor);
}

void usage()
{
        printf("usage: aftp [options] <value>\n");
        printf("options are:\n");
        printf("        -h<help >\n");
        printf("        -f<File Size in Bytes>\n");
        printf("        -b<Bit rate in Kbps>\n");
        printf("        -p<Packet Size in Bytes>\n");
        printf("        -r<No of Receivers>\n");
        exit(0);
}


main(int argc, char *argv[])
{

    if (argc < 2) {
       usage();
       exit(1);
    }

    argc--; argv++;
    while (argc) {
	    if (argv[0][0] != '-')
		    break;
	    switch (argv[0][1]) {

	    case 'h':
	            usage();
		    break;
	    case 'f':
		    sscanf(&argv[0][2], "%d", &fileSize);
		    printf("  File size (Bytes): %d\n", fileSize);
		    break;
	    case 'p':
		    sscanf(&argv[0][2], "%d", &MSize);
		    printf("Packet size (Bytes): %d\n", MSize);
		    break;
	    case 'b':
		    sscanf(&argv[0][2], "%d", &rate);
		    printf("   Data Rate (Kbps): %d\n", rate);
		    break;
	    case 'r':
		    sscanf(&argv[0][2], "%d", &nReceivers);
		    printf("Number of Receivers: %d\n", nReceivers);
		    break;
	    default:
		    printf("Error: invalid option\n");
		    usage();
	    }
	    argc--; argv++;
    }


// Set the SEED value based on the system time
if (first_time) {
    first_time = FALSE;
    srand((unsigned int) time (NULL));
}  

// Get the fade values
fade[0]=100.0;
fade[1]=70.0;
fade[2]=20.0;
fade[3]=10.0;
fade[4]=6.0;
fade[5]=4.0;
fade[6]=2.75;
fade[7]=2.0;
fade[8]=1.75;
fade[9]=1.5;
fade[10]=1.0;

// Compute the transfer time
file_transfer_time = ((double)fileSize * 8.0) /((double)rate*1000);

// Compute Number of packets
nPackets = (unsigned long)ceil(((double)fileSize)/(double)MSize);

printf("Scanning %d Receivers.....\n", nReceivers);
for (j = 1; j <= nReceivers; j++) {

     // initialise performance parameters
     lost_packets[j] = 0;
     duplicates[1][j] = 0;
     recvdPackets[1][j] = 0;
     usefulPackets[1][j] = 0;
     nPass[j] = 1;
     start_fade_time = 0.0;
     end_fade_time = 0.0;
     end_phase_time = 0.0;
     lost_pkts_first = 0;
     lost_pkts_last = 0;

     // Set the end of phase 
     end_phase_time = (double)file_transfer_time;

     // Choose fade duration from uniform distribution
     char * s;
     sprintf(s, "%3.1f", 10*uniform());
     int d = atoi(s);
     //     fade_time = fade[d];

     if (j == 1) {
       fade_time = fade[1];
       start_fade_time = 6.0;
     }

     if (j == 2) {
       fade_time = fade[3];
       start_fade_time = 4.0;
     }
       
     
     if (j == 3) {  
       fade_time = fade[6];
       start_fade_time = 5.0;
     }   

     if (j == 4) {
       fade_time = fade[8];
       start_fade_time = 9.0; 
     }

     // printf("fade time is: %f and end phase time is: %f\n", fade_time,end_phase_time);

     // Choose a random start time for the fade & set the end of fade time
     // start_fade_time = (double)(rand()%(2*(int)(ceil(file_transfer_time))));
     end_fade_time = (start_fade_time + fade_time);
     
     // printf("start fade time is: %f end fade time is: %f\n", start_fade_time,end_fade_time);

     packets = 0;
     while (packets < nPackets) {

     // check for packets in the 1st phase
     if (nPass[j] == 1) {

       // check for lost packets when there is no fade
       if (start_fade_time >= end_phase_time) {
	 lost_packets[j] = 0;
	 duplicates[nPass[j]][j] = 0;
         recvdPackets[nPass[j]][j] = nPackets;
         usefulPackets[nPass[j]][j] = nPackets;
         loss[nPass[j]][j] = lost_packets[j];
         packets = nPackets;
       }

       // check for lost packets when fade ends in the first phase
       else if (end_fade_time < end_phase_time) {
	  lost_packets[j] = (unsigned long)((rate*1000*fade_time)/(8*MSize));
          usefulPackets[nPass[j]][j] = (nPackets - lost_packets[j]);
          recvdPackets[nPass[j]][j] = usefulPackets[nPass[j]][j];
          loss[nPass[j]][j] = lost_packets[j];
         
          nPass[j]=(nPass[j]+1);
          usefulPackets[nPass[j]][j] = lost_packets[j];
          duplicates[nPass[j]][j] = (nPackets - lost_packets[j]);
          recvdPackets[nPass[j]][j] = nPackets;
          loss[nPass[j]][j] = 0;     
	  packets = nPackets;
       }

       // check for lost packets when fade continues to second phase
       else if ((start_fade_time < end_phase_time) && (end_fade_time > end_phase_time)) {
	  lost_packets[j] = (unsigned long)((rate*1000*(end_phase_time-start_fade_time))/(8*MSize));
          lost_pkts_first = lost_packets[j];

  	  usefulPackets[nPass[j]][j] = (nPackets - lost_packets[j]);
          recvdPackets[nPass[j]][j] = usefulPackets[nPass[j]][j];
          loss[nPass[j]][j] = lost_packets[j];
          nPass[j]=(nPass[j]+1);

  	  packets = (nPackets - lost_packets[j]);
 	  start_fade_time = end_phase_time;
	  end_phase_time = (end_phase_time + file_transfer_time);
       }
     }
     else
     {
       // check for lost packets in the subsequent phases
       if (end_fade_time > end_phase_time) {
	  lost_packets[j] = (unsigned long)((rate*1000*(end_phase_time - start_fade_time))/(8*MSize))+lost_packets[j];
  	  usefulPackets[nPass[j]][j] = 0;
          duplicates[nPass[j]][j] = 0;
          recvdPackets[nPass[j]][j] = usefulPackets[nPass[j]][j];
          loss[nPass[j]][j] =  (unsigned long)((rate*1000*(end_phase_time - start_fade_time))/(8*MSize));
          nPass[j]=(nPass[j]+1);

	  start_fade_time = end_phase_time;
	  end_phase_time = (end_phase_time + file_transfer_time);	
       }

       // check for lost packets in the second phase
       else if (end_fade_time <= end_phase_time) {
	  lost_pkts_last = (unsigned long)((rate*1000*(end_fade_time - start_fade_time))/(8*MSize));
          lost_packets[j] = lost_packets[j]+lost_pkts_last;
          tot_lost_pkts = (lost_pkts_first+lost_pkts_last);

  	  usefulPackets[nPass[j]][j] = (nPackets - lost_pkts_last);
          recvdPackets[nPass[j]][j] = usefulPackets[nPass[j]][j];
          loss[nPass[j]][j] =  lost_pkts_last;

          if (tot_lost_pkts <= nPackets) {
             duplicates[nPass[j]][j] = (nPackets - tot_lost_pkts);
          }
          else {
             duplicates[nPass[j]][j] = 0;
	  }

          nPass[j]=(nPass[j]+1);
  	  usefulPackets[nPass[j]][j] = min(lost_pkts_first,lost_pkts_last);
	  //          printf("minimum of %d and %d is: %d\n",lost_pkts_first,lost_pkts_last,min(lost_pkts_first,lost_pkts_last));
          duplicates[nPass[j]][j] = (nPackets - usefulPackets[nPass[j]][j]);
          recvdPackets[nPass[j]][j] = nPackets;
          loss[nPass[j]][j] =  0;

          packets = nPackets;
       }
     } // end if  
   } // end while

   // Get the maximum number of passes
     //     printf("nPass is: %d\n",nPass[j]);
   if (nPass[j] > maxPass) {
     maxPass = nPass[j];
   }
 } // end for
 
 // output the values of Packet Loss 

  fprintf(fd_cumlossrate, "\nAFTP Simulation data\n");
  fprintf(fd_cumlossrate, "Receivers: %d, Pkt Size: %d Bytes, File Size: %d Bytes, Rate: %d Kbps\n", nReceivers,MSize,fileSize,rate);
  fprintf(fd_cumlossrate, "Time(s) Cumulative Loss Rate\n\n");
  fprintf(fd_loss, "\nAFTP Simulation data\n");
  fprintf(fd_loss, "Receivers: %d, Pkt Size: %d Bytes, File Size: %d Bytes, Rate: %d Kbps\n", nReceivers,MSize,fileSize,rate);
  fprintf(fd_loss, "Time(s) Loss Rate\n\n");
  fprintf(fd_goodput, "\nAFTP Simulation data\n");
  fprintf(fd_goodput, "Receivers: %d, Pkt Size: %d Bytes, File Size: %d Bytes, Rate: %d Kbps\n", nReceivers,MSize,fileSize,rate);
  fprintf(fd_goodput, "Time(s) Goodput\n\n");
  fprintf(fd_throughput, "\nAFTP Simulation data\n");
  fprintf(fd_throughput, "Receivers: %d, Pkt Size: %d Bytes, File Size: %d Bytes, Rate: %d Kbps\n", nReceivers,MSize,fileSize,rate);
  fprintf(fd_throughput, "Time(s) Throughput\n\n");
  fprintf(fd_bandwidth, "\nAFTP Simulation data\n");
  fprintf(fd_bandwidth, "Receivers: %d, Pkt Size: %d Bytes, File Size: %d Bytes, Rate: %d Kbps\n", nReceivers,MSize,fileSize,rate);
  fprintf(fd_bandwidth, "Time(s) Bandwidth\n\n");
  fprintf(fd_dups,"\nTime (s)  Total Duplicates - Receivers: %d File Size: %d\n",nReceivers,fileSize);

  totalPass = 0;

  for (j = 1; j <= nReceivers; j++) {
      fprintf(fd_goodput,"\nReceiver: %d\n",j);
      fprintf(fd_loss,"\nReceiver: %d\n",j);
      fprintf(fd_cumlossrate,"\nReceiver: %d\n",j);      
      fprintf(fd_throughput, "\nReceiver: %d\n",j);
      printf("reciver: %d pass: %d\n",j,nPass[j]);

      sumLoss = 0;
      for (k = 1; k <= nPass[j]; k++) {
 	  sumLoss = (sumLoss + loss[k][j]);
          cumlossrate[k][j] = ((double)sumLoss/(double)(k*nPackets))*100;
          lossrate[k][j] = (double)loss[k][j]/(double)nPackets;
	  goodput[k][j] = (double)recvdPackets[k][j]/(double)nPackets; 
          throughput[k][j] = (double)(recvdPackets[k][j]*MSize*8)/(double)(file_transfer_time*1000);
          bandwidth[k][j] = (double)(nPackets)/(double)(recvdPackets[k][j]);
          
          printf("pass: %d receiver: %d goodput: %f received: %d\n",k,j,goodput[k][j],usefulPackets[k][j]);
          
          fprintf(fd_loss,"%f  %f\n",(k*file_transfer_time),lossrate[k][j]);
          fprintf(fd_cumlossrate,"%f  %f\n",(k*file_transfer_time),cumlossrate[k][j]);
          fprintf(fd_goodput,"%f  %f\n",(k*file_transfer_time),goodput[k][j]);         
          fprintf(fd_throughput,"%f %f\n",(k*file_transfer_time),throughput[k][j]);
          fprintf(fd_bandwidth,"%f %f\n",(k*file_transfer_time),bandwidth[k][j]);
      }
  }

  printf("Successfully Completed...\n\n");

  fclose(fd_loss);
  fclose(fd_cumlossrate);
  fclose(fd_goodput);
  fclose(fd_throughput);
  fclose(fd_bandwidth);
}





