# (c) 1999-2000 Aberdeen University Electronics Research Lab
#
# THIS SOFTWARE IS PROVIDED BY THE CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
#
#
#    Designed by: Martin Koyabe, koyabe@erg.abdn.ac.uk
#           File: cbr2_sim_general.tcl
#    Last update: 22 December 2000

# This is a baseline script that creates (1) source and (N) receiver nodes with asymmetric links.
# Each downlink experiences both random packet errors and losses (generated by error model). The 
# source generates CBR UDP packets of length 1472 Bytes to the multicast group. Receivers register
# randomly to start receiving packets



# setup simulator
set ns [new Simulator -multicast on]


# create the trace-all output file
set f1 [open aftp_sim_general_1.tr w]
set f2 [open aftp_sim_general_2.tr w]
set f3 [open aftp_sim_general_3.tr w]
set f4 [open aftp_sim_general_4.tr w]
set f5 [open aftp_sim_general_5.tr w]
set f6 [open aftp_sim_general_6.tr w]
set f7 [open aftp_sim_general_7.tr w]
set f8 [open aftp_sim_general_8.tr w]
set f9 [open aftp_sim_general_9.tr w]
set f10 [open aftp_sim_general_10.tr w]
set f11 [open aftp_sim_general_11.tr w]
set f12 [open aftp_sim_general_12.tr w]


# create the nam output file
#set nf [open cbr_sim_general.nam w]
#$ns namtrace-all $nf


set rng [new RNG]
$rng seed 0
        
set range [list 0 100]
set start_range [list 6.0 4.0 5.0 9.0]
set duration_range [list 70.0 10.0 2.75 0.0]
set time_range [list 0 0 0 0]

for {set i 0} {$i < 4} {incr i} {
    set f_start($i) [list [lindex $start_range $i]]
    set s_time($i) [lindex $time_range $i]
    set f_duration($i) [list [lindex $duration_range $i]]
}

# set s_time 3

for {set k 0} {$k < [expr (5-1)]} {incr k} {

#     set s_time($k) [expr [expr [$rng next-random]] % [lindex $range 1]]  
#    while {$s_time($k) < [lindex $range 0]} {  
#          set s_time($k) [expr [expr [$rng next-random]] % [lindex $range 1]]    
#    }

    set error_model_($k) [new ErrorModel/TimeSeries $f_duration($k) $f_start($k) $range [expr [lindex $time_range $k]]]
    $error_model_($k) unit pkt
    $error_model_($k) set rate_ 1.0
}

# Create Error Model

#for {set k 0} {$k < [expr (4-2)]} {incr k} {
#    set error_model($k) [new ErrorModel/]
#    $error_model($k) unit pkt
#    $error_model($k) set rate_ 0.5
#    $error_model($k) ranvar [new RandomVariable/Uniform]
#    $error_model($k) drop-target [new Agent/Null]
#}


# To create a Random Multi-state fade error model, the following parameters should 
# be supplied (as defined in /tcl/lib/ns-errmodel.tcl): 

#       states: an array of states (error models). 
#      periods: an array of average state durations. 
#  probability: an array of state probability.
#    transunit: one of [pkt|byte|time]. 
#       sttype: type of state transitions to use: either time or pkt. 


# set tmp  [new ErrorModel/Uniform 0.000 bit]
# set tmp1 [new ErrorModel/Uniform 1.000 bit]
# set tmp2 [new ErrorModel/Uniform 0.5 pkt]
 
# array of states (error models)
# set m_states [list $tmp $tmp1]

# average fade periods for each of the states, tmp, tmp1 and tmp2, respectively
# set m_periods [list 150 32]

# probability distribution of fades
# set m_probability [list 0.95 0.05]

# set m_trunit pkt

# Use time-based transition
# set m_sttype time

# assign error model to a variable
# for {set k 0} {$k < 3} {incr k} {
#        set error_model_($k) [new ErrorModel/RandomMultiState $m_states $m_periods $m_probability $m_trunit $m_sttype]
# }

# create network topology like this:
#                          2
#                         /.
#                        / .
#              0--------1  .
#                        \ .
#                         \.
#                          n


# setup nodes

for {set i 0} {$i < 6} {incr i} {
      set node_($i) [$ns node]
}


# Creating links

$ns duplex-link $node_(0) $node_(1)  1.0mb 0ms DropTail
$ns duplex-link $node_(1) $node_(2)  1.0mb 0ms DropTail
$ns duplex-link $node_(1) $node_(3)  1.0mb 0ms DropTail
$ns duplex-link $node_(1) $node_(4)  1.0mb 0ms DropTail
$ns duplex-link $node_(1) $node_(5)  1.0mb 0ms DropTail

# setting the queue limit

$ns queue-limit $node_(0) $node_(1)  10000000
$ns queue-limit $node_(1) $node_(0)  10000000
$ns queue-limit $node_(1) $node_(2)  10000000
$ns queue-limit $node_(2) $node_(1)  10000000
$ns queue-limit $node_(1) $node_(3)  10000000
$ns queue-limit $node_(3) $node_(1)  10000000
$ns queue-limit $node_(1) $node_(4)  10000000
$ns queue-limit $node_(4) $node_(1)  10000000
$ns queue-limit $node_(1) $node_(5)  10000000
$ns queue-limit $node_(5) $node_(1)  10000000


# inserting error model along the links

$ns link-lossmodel $error_model_(0) $node_(1) $node_(2)
$ns link-lossmodel $error_model_(1) $node_(1) $node_(3)
$ns link-lossmodel $error_model_(2) $node_(1) $node_(4)
$ns link-lossmodel $error_model_(3) $node_(1) $node_(5) 

# Set the routing protocol

$ns mrtproto DM

# Allocate group address

set group [Node allocaddr]


# Transport agent for the traffic source

set udp0 [new Agent/UDP]
$ns attach-agent $node_(0) $udp0
$udp0 set dst_addr_ $group
$udp0 set dst_port_ 1



# Set the constant bit rate from source 0

set cbr0 [new Application/Traffic/CBR]
$cbr0 attach-agent $udp0
$udp0 set packetSize_ 1472
$udp0 set rate_ 1000Kb


# start the constant bit rate

$ns at 0.0 "$cbr0 start"

# Set the receiver agent
# Can also be Agent/Null


#for {set j 0} {$j < [expr (4-2)]} {incr j} {
#    set rcvr_($j) [new Agent/LossMonitor]
#    $ns attach-agent $node_([expr ($j+2)]) $rcvr_($j)
#}


    set rcvr0 [new Agent/LossMonitor]
    $ns attach-agent $node_(2) $rcvr0

    set rcvr1 [new Agent/LossMonitor]
    $ns attach-agent $node_(3) $rcvr1

    set rcvr2 [new Agent/LossMonitor]
    $ns attach-agent $node_(4) $rcvr2

    set rcvr3 [new Agent/LossMonitor]
    $ns attach-agent $node_(5) $rcvr3



# Start the record procedure

$ns at 8.192 "record"

# Join the group at some point

$ns at 0.0 "$node_(2) join-group  $rcvr0 $group"
$ns at 0.0 "$node_(3) join-group  $rcvr1 $group"
$ns at 0.0 "$node_(4) join-group  $rcvr2 $group"
$ns at 0.0 "$node_(5) join-group  $rcvr3 $group"
 
# $ns at 200.0 "$cbr0 stop"

$ns at 100 "$node_(2) leave-group $rcvr0 $group"
$ns at 100  "$node_(3) leave-group $rcvr1 $group"
$ns at 100  "$node_(4) leave-group $rcvr2 $group"
$ns at 100  "$node_(5) leave-group $rcvr3 $group"

puts "Transmitting Packets ..."; flush stdout
puts "simulation is running..."; flush stdout

$ns at 100.0 "finish"
         
          set totallost0 0
          set totallost1 0
          set totallost2 0
          set totallost3 0
    
          set totalpkts0 0
          set totalpkts1 0
          set totalpkts2 0
          set totalpkts3 0

          set totalrcvd0 0
          set totalrcvd1 0
          set totalrcvd2 0
          set totalrcvd3 0

          set lost0 0
          set lost1 0
          set lost2 0
          set lost3 0

          set pkts0 0
          set pkts1 0
          set pkts2 0
          set pkts3 0

          set bytes0 0
          set bytes1 0
          set bytes2 0
          set bytes3 0

          set count 1

          set real_loss0 0
          set real_loss1 0
          set real_loss2 0
          set real_loss3 0

proc record  { } {
           global count rcvr0 rcvr1 rcvr2 rcvr3 f1 f2 f3 f4 f5 f6 f7 f8 f9 f10 f11 f12 totallost0 totallost1 totallost2 totallost3  
           global totalpkts0 totalpkts1 totalpkts2 totalpkts3 totalrcvd0 totalrcvd1 totalrcvd2 totalrcvd3       
           
           #Get an instance of the simulator
           set ns [Simulator instance]
           
           #Set the time after which the procedure should be called again
           set time 8.192


           #compute the number of packets sent
           set nPackets ceil([expr [expr 1000000*$time]/[expr 8*1472]])

           #How many bytes have been received by the traffic sinks?

           set bytes0 [$rcvr0 set bytes_]
           set bytes1 [$rcvr1 set bytes_]
           set bytes2 [$rcvr2 set bytes_]
           set bytes3 [$rcvr3 set bytes_]

           set pkts0 [$rcvr0 set npkts_]
           set pkts1 [$rcvr1 set npkts_]
           set pkts2 [$rcvr2 set npkts_]
           set pkts3 [$rcvr3 set npkts_]

           set totalrcvd0 ceil([expr [expr 1000000*$time]/[expr 8*1472]])
           set totalrcvd1 ceil([expr [expr 1000000*$time]/[expr 8*1472]])
           set totalrcvd2 ceil([expr [expr 1000000*$time]/[expr 8*1472]])
           set totalrcvd3 ceil([expr [expr 1000000*$time]/[expr 8*1472]])
 
           set lost0 [expr $totalrcvd0 - $pkts0]
           set lost1 [expr $totalrcvd1 - $pkts1]
           set lost2 [expr $totalrcvd2 - $pkts2]
           set lost3 [expr $totalrcvd3 - $pkts3]

           #Get the current time
           set now [$ns now]

           puts "Receiver 1 - pkts lost: $lost0 pkts received: $pkts0 total received: $totalrcvd0"
           puts "Receiver 2 - pkts lost: $lost1 pkts received: $pkts1 total received: $totalrcvd1"

             
           # Calculate the bandwidth (in KBit/s) and write it to the files

	   puts $f1 "$now  [expr $bytes0/$time*8/1000]" 
           puts $f2 "$now  [expr $bytes1/$time*8/1000]"
           puts $f3 "$now  [expr $bytes2/$time*8/1000]"
           puts $f4 "$now  [expr $bytes3/$time*8/1000]"


           # Calculate the goodput ratio

	   puts $f5 "$now  [expr $pkts0-$real_loss0/$totalrcvd0]" 
	   puts $f6 "$now  [expr $pkts1/$totalrcvd1]" 
	   puts $f7 "$now  [expr $pkts2/$totalrcvd2]" 
	   puts $f8 "$now  [expr $pkts3/$totalrcvd3]" 


           # Calculate the cumulative loss rate 
   
           if { $totalrcvd0 != 0 } { puts $f9  "$now  [format "%5.3f" [expr double($lost0)/double($totalrcvd0)]]" 
              puts "loss rate at: $now is: [expr $lost0/$totalrcvd0]"
           }
           if { $totalrcvd1 != 0 } { puts $f10 "$now  [format "%5.3f" [expr double($lost1)/double($totalrcvd1)]]" }
           if { $totalrcvd2 != 0 } { puts $f11 "$now  [format "%5.3f" [expr double($lost2)/double($totalrcvd2)]]" }
           if { $totalrcvd3 != 0 } { puts $f12 "$now  [format "%5.3f" [expr double($lost3)/double($totalrcvd3)]]" }


           #Reset the bytes_ values on the traffic sinks

           $rcvr0 set bytes_ 0
           $rcvr1 set bytes_ 0
           $rcvr2 set bytes_ 0
           $rcvr3 set bytes_ 0

#           $rcvr0 set nlost_ 0
#           $rcvr1 set nlost_ 0
#           $rcvr2 set nlost_ 0
#           $rcvr3 set nlost_ 0

           $rcvr0 set npkts_ 0
           $rcvr1 set npkts_ 0
           $rcvr2 set npkts_ 0
           $rcvr3 set npkts_ 0

           #Re-schedule the procedure

           $ns at [expr $now+$time] "record"
           set count [expr $count+1]
}

proc finish { } {
    global ns nf f1 f2 f3 f4 f5 f6 f7 f8 rcvr0 rcvr1 rcvr2 rcvr3 

#    for {set k 0} {$k <  [expr (4-2)]} {incr k} {
        puts "receiver 0 --"
        puts "lossMonitor nlost: [$rcvr0 set nlost_]";                   # Number of packets lost
        puts "lossMonitor nreceived: [$rcvr0 set npkts_]";               # Number of packets received 
        puts "lossMonitor expected: [$rcvr0 set expected_]";             # The expected sequence number of the next pkt
        puts "lossMonitor last packet time: [$rcvr0 set lastPktTime_]";  # Time at which the last packet was received
        puts "lossMonitor bytes: [$rcvr0 set bytes_]";                   # Number of bytes received
        puts " "

        puts "receiver 1 --"
        puts "lossMonitor nlost: [$rcvr1 set nlost_]" 
        puts "lossMonitor nreceived: [$rcvr1 set npkts_]"
        puts "lossMonitor expected: [$rcvr1 set expected_]"
        puts "lossMonitor last packet time: [$rcvr1 set lastPktTime_]"
        puts "lossMonitor bytes: [$rcvr1 set bytes_]"
        puts " "

        puts "receiver 2 --"
        puts "lossMonitor nlost: [$rcvr2 set nlost_]" 
        puts "lossMonitor nreceived: [$rcvr2 set npkts_]"
        puts "lossMonitor expected: [$rcvr2 set expected_]"
        puts "lossMonitor last packet time: [$rcvr2 set lastPktTime_]"
        puts "lossMonitor bytes: [$rcvr2 set bytes_]"
        puts " "

        puts "receiver 3 --"
        puts "lossMonitor nlost: [$rcvr3 set nlost_]" 
        puts "lossMonitor nreceived: [$rcvr3 set npkts_]"
        puts "lossMonitor expected: [$rcvr3 set expected_]"
        puts "lossMonitor last packet time: [$rcvr3 set lastPktTime_]"
        puts "lossMonitor bytes: [$rcvr3 set bytes_]"
        puts " "


#    }

    $ns flush-trace
#    close $nf
    puts stdout "view output in cbr_sim_general.tr"
    puts stdout "Simulation successfully completed ...."
    puts stdout "Now Running Nam ..."
    exit 0
    exec xgraph cbr_sim_general_1.tr cbr_sim_general_2.tr &
}

$ns run








